from pathlib import Path
import re
from typing import Sequence, Union

import aiohttp
from tqdm import tqdm

from .base import BaseFunction, SyncFunctionMixin
from .config import APIConfig
from .request import Request
from .cli.pretty import ProgressReportingReader

__all__ = (
    'BaseVFolder',
    'VFolder',
)

_rx_slug = re.compile(r'^[a-zA-Z0-9]([a-zA-Z0-9._-]*[a-zA-Z0-9])?$')


class BaseVFolder(BaseFunction):
    @classmethod
    def _create(cls, name: str, *,
                config: APIConfig=None):
        assert _rx_slug.search(name) is not None
        resp = yield Request('POST', '/folders/', {
            'name': name,
        }, config=config)
        return resp.json()

    @classmethod
    def _list(cls, *, config: APIConfig=None):
        resp = yield Request('GET', '/folders/', config=config)
        return resp.json()

    @classmethod
    def _get(cls, name: str, *, config: APIConfig=None):
        return cls(name, config=config)

    def _info(self):
        resp = yield Request('GET', '/folders/{0}'.format(self.name),
                             config=self.config)
        return resp.json()

    def _delete(self):
        resp = yield Request('DELETE', '/folders/{0}'.format(self.name),
                             config=self.config)
        if resp.status == 200:
            return resp.json()

    def _upload(self, files: Sequence[Union[str, Path]],
               basedir: Union[str, Path]=None,
               show_progress: bool=False):
        fields = []
        base_path = (Path.cwd() if basedir is None
                     else Path(basedir).resolve())
        files = [Path(file).resolve() for file in files]
        total_size = 0
        for file_path in files:
            total_size += file_path.stat().st_size
        tqdm_obj = tqdm(desc='Uploading files',
                        unit='bytes', unit_scale=True,
                        ncols=79,
                        total=total_size,
                        disable=not show_progress)
        with tqdm_obj:
            for file_path in files:
                try:
                    fields.append(aiohttp.web.FileField(
                        'src',
                        str(file_path.relative_to(base_path)),
                        ProgressReportingReader(str(file_path),
                                                tqdm_instance=tqdm_obj),
                        'application/octet-stream',
                        None
                    ))
                except ValueError:
                    msg = 'File "{0}" is outside of the base directory "{1}".' \
                          .format(file_path, base_path)
                    raise ValueError(msg) from None
            rqst = Request('POST', '/folders/{}/upload'.format(self.name),
                           config=self.config,
                           reporthook=None)
            rqst.content = fields
            resp = yield rqst
        return resp

    def _download(self, files: Sequence[Union[str, Path]]):
        # TODO: implement
        raise NotImplementedError

    def __init__(self, name: str, *, config: APIConfig=None):
        assert _rx_slug.search(name) is not None
        self.name = name
        self.config = config
        self.delete   = self._call_base_method(self._delete)
        self.info     = self._call_base_method(self._info)
        self.upload   = self._call_base_method(self._upload)
        self.download = self._call_base_method(self._download)

    def __init_subclass__(cls):
        cls.create = cls._call_base_clsmethod(cls._create)
        cls.list   = cls._call_base_clsmethod(cls._list)
        cls.get    = cls._call_base_clsmethod(cls._get)


class VFolder(SyncFunctionMixin, BaseVFolder):
    pass



from typing import Iterable, Mapping, Sequence, Union
from pathlib import Path
import uuid

import aiohttp.web
from tqdm import tqdm

from .base import BaseFunction, SyncFunctionMixin
from .config import APIConfig, get_config
from .exceptions import BackendClientError
from .request import Request
from .cli.pretty import ProgressReportingReader

__all__ = (
    'BaseKernel',
    'Kernel',
)


class BaseKernel(BaseFunction):

    '''
    Implements the request creation and response handling logic,
    while delegating the process of request sending to the subclasses
    via the generator protocol.
    '''

    @classmethod
    def _get_or_create(cls, lang: str,
                       client_token: str=None,
                       mounts: Iterable[str]=None,
                       envs: Mapping[str, str]=None,
                       max_mem: int=0, exec_timeout: int=0,
                       config: APIConfig=None) -> str:
        if client_token:
            assert 4 <= len(client_token) <= 64, \
                   'Client session token should be 4 to 64 characters long.'
        else:
            client_token = uuid.uuid4().hex
        if config is None:
            config = get_config()
        if mounts is None:
            mounts = []
        mounts.extend(config.vfolder_mounts)
        resp = yield Request('POST', '/kernel/create', {
            'lang': lang,
            'clientSessionToken': client_token,
            'config': {
                'mounts': mounts,
                'environ': envs,
            },
        }, config=config)
        data = resp.json()
        o = cls(data['kernelId'], config=config)  # type: ignore
        o.created = data.get('created', True)     # True is for legacy
        return o

    def _destroy(self):
        resp = yield Request('DELETE', '/kernel/{}'.format(self.kernel_id),
                             config=self.config)
        if resp.status == 200:
            return resp.json()

    def _restart(self):
        yield Request('PATCH', '/kernel/{}'.format(self.kernel_id),
                      config=self.config)

    def _interrupt(self):
        yield Request('POST', '/kernel/{}/interrupt'.format(self.kernel_id),
                      config=self.config)

    def _complete(self, code: str, opts: dict=None):
        opts = {} if opts is None else opts
        rqst = Request('POST', '/kernel/{}/complete'.format(self.kernel_id), {
            'code': code,
            'options': {
                'row': int(opts.get('row', 0)),
                'col': int(opts.get('col', 0)),
                'line': opts.get('line', ''),
                'post': opts.get('post', ''),
            },
        }, config=self.config)
        resp = yield rqst
        return resp.json()

    def _get_info(self):
        resp = yield Request('GET', '/kernel/{}'.format(self.kernel_id),
                             config=self.config)
        return resp.json()

    def _get_logs(self):
        resp = yield Request('GET', '/kernel/{}/logs'.format(self.kernel_id),
                             config=self.config)
        return resp.json()

    def _execute(self, run_id: str=None,
                 code: str=None,
                 mode: str='query',
                 opts: dict=None):
        opts = {} if opts is None else opts
        if mode in {'query', 'continue', 'input'}:
            assert code is not None  # but maybe empty due to continuation
            rqst = Request('POST', '/kernel/{}'.format(self.kernel_id), {
                'mode': mode,
                'code': code,
                'runId': run_id,
            }, config=self.config)
        elif mode == 'batch':
            rqst = Request('POST', '/kernel/{}'.format(self.kernel_id), {
                'mode': mode,
                'code': code,
                'runId': run_id,
                'options': {
                    'build': opts.get('build', None),
                    'buildLog': bool(opts.get('buildLog', False)),
                    'exec': opts.get('exec', None),
                },
            }, config=self.config)
        elif mode == 'complete':
            rqst = Request('POST', '/kernel/{}/complete'.format(self.kernel_id), {
                'code': code,
                'options': {
                    'row': int(opts.get('row', 0)),
                    'col': int(opts.get('col', 0)),
                    'line': opts.get('line', ''),
                    'post': opts.get('post', ''),
                },
            }, config=self.config)
        else:
            raise BackendClientError('Invalid execution mode: {0}'.format(mode))
        resp = yield rqst
        return resp.json()['result']

    def _upload(self, files: Sequence[Union[str, Path]],
               basedir: Union[str, Path]=None,
               show_progress: bool=False):
        fields = []
        base_path = (Path.cwd() if basedir is None
                     else Path(basedir).resolve())
        files = [Path(file).resolve() for file in files]
        total_size = 0
        for file_path in files:
            total_size += file_path.stat().st_size
        tqdm_obj = tqdm(desc='Uploading files',
                        unit='bytes', unit_scale=True,
                        ncols=79,
                        total=total_size,
                        disable=not show_progress)
        with tqdm_obj:
            for file_path in files:
                try:
                    fields.append(aiohttp.web.FileField(
                        'src',
                        str(file_path.relative_to(base_path)),
                        ProgressReportingReader(str(file_path),
                                                tqdm_instance=tqdm_obj),
                        'application/octet-stream',
                        None
                    ))
                except ValueError:
                    msg = 'File "{0}" is outside of the base directory "{1}".' \
                          .format(file_path, base_path)
                    raise ValueError(msg) from None
            rqst = Request('POST', '/kernel/{}/upload'.format(self.kernel_id),
                           config=self.config)
            rqst.content = fields
            resp = yield rqst
        return resp

    def __init__(self, kernel_id: str, *, config: APIConfig=None) -> None:
        self.kernel_id = kernel_id
        self.config    = config
        self.destroy   = self._call_base_method(self._destroy)
        self.restart   = self._call_base_method(self._restart)
        self.interrupt = self._call_base_method(self._interrupt)
        self.complete  = self._call_base_method(self._complete)
        self.get_info  = self._call_base_method(self._get_info)
        self.get_logs  = self._call_base_method(self._get_logs)
        self.execute   = self._call_base_method(self._execute)
        self.upload    = self._call_base_method(self._upload)

    def __init_subclass__(cls):
        cls.get_or_create = cls._call_base_clsmethod(cls._get_or_create)


class Kernel(SyncFunctionMixin, BaseKernel):
    pass



import os
from yarl import URL
from typing import Iterable, Tuple, Union

__all__ = [
    'APIConfig',
]

_config = None


def get_env(name, default=None, clean=lambda v: v):
    v = os.environ.get('BACKEND_' + name)
    if v is None:
        v = os.environ.get('SORNA_' + name)
    if v is None:
        if default is None:
            raise KeyError(name)
        v = default
    return clean(v)


def _clean_url(v):
    v = v if isinstance(v, URL) else URL(v)
    if not v.is_absolute():
        raise ValueError('URL must be absolute.')
    return v


def _clean_tokens(v):
    if isinstance(v, str):
        if not v:
            return tuple()
        return tuple(v.split(','))
    return tuple(iter(v))


class APIConfig:

    DEFAULTS = {
        'endpoint': 'https://api.backend.ai',
        'version': 'v2.20170315',
        'hash_type': 'sha256',
        'vfolder_mounts': [],
    }

    def __init__(self, *,
                 endpoint: Union[URL, str]=None,
                 version: str=None,
                 user_agent: str=None,
                 access_key: str=None,
                 secret_key: str=None,
                 hash_type: str=None,
                 vfolder_mounts: Iterable[str]=None) -> None:
        from . import get_user_agent  # noqa; to avoid circular imports
        self._endpoint = (
            _clean_url(endpoint) if endpoint else
            get_env('ENDPOINT', self.DEFAULTS['endpoint'], clean=_clean_url))
        self._version = version if version else self.DEFAULTS['version']
        self._user_agent = user_agent if user_agent else get_user_agent()
        self._access_key = access_key if access_key else get_env('ACCESS_KEY')
        self._secret_key = secret_key if secret_key else get_env('SECRET_KEY')
        self._hash_type = hash_type.lower() if hash_type else \
                          self.DEFAULTS['hash_type']
        self._vfolder_mounts = (
            _clean_tokens(vfolder_mounts) if vfolder_mounts else
            get_env('VFOLDER_MOUNTS', self.DEFAULTS['vfolder_mounts'],
                    clean=_clean_tokens))

    @property
    def endpoint(self) -> URL:
        return self._endpoint

    @property
    def user_agent(self) -> str:
        return self._user_agent

    @property
    def access_key(self) -> str:
        return self._access_key

    @property
    def secret_key(self) -> str:
        return self._secret_key

    @property
    def version(self) -> str:
        return self._version

    @property
    def hash_type(self) -> str:
        return self._hash_type

    @property
    def vfolder_mounts(self) -> Tuple[str, ...]:
        return self._vfolder_mounts


def get_config():
    global _config
    if _config is None:
        _config = APIConfig()
    return _config


def set_config(conf: APIConfig):
    global _config
    _config = conf



import asyncio
from collections import OrderedDict
from datetime import datetime
import queue
import threading
from typing import Any, Callable, Mapping, Sequence, Union

import aiohttp
import aiohttp.web
from async_timeout import timeout as _timeout
from dateutil.tz import tzutc
from multidict import CIMultiDict
import json

from .auth import generate_signature
from .config import APIConfig, get_config
from .exceptions import BackendClientError

__all__ = [
    'Request',
    'Response',
]

_worker_thread = None


class _SyncWorkerThread(threading.Thread):

    sentinel = object()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.work_queue = queue.Queue()
        self.done_queue = queue.Queue()

    def run(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            while True:
                coro = self.work_queue.get()
                if coro is self.sentinel:
                    break
                try:
                    result = loop.run_until_complete(coro)
                except Exception as e:
                    self.done_queue.put_nowait(e)
                else:
                    self.done_queue.put_nowait(result)
                self.work_queue.task_done()
        except (SystemExit, KeyboardInterrupt):
            pass
        finally:
            loop.close()


def shutdown():
    global _worker_thread
    if _worker_thread is not None:
        _worker_thread.work_queue.put(_worker_thread.sentinel)
        _worker_thread.join()
        _worker_thread = None


class BaseRequest:

    __slots__ = ['config', 'method', 'path',
                 'date', 'headers',
                 'content_type', '_content']

    _allowed_methods = frozenset([
        'GET', 'HEAD', 'POST',
        'PUT', 'PATCH', 'DELETE',
        'OPTIONS'])

    def __init__(self, method: str='GET',
                 path: str=None,
                 content: Mapping=None,
                 config: APIConfig=None,
                 reporthook: Callable=None) -> None:
        '''
        Initialize an API request.

        :param str path: The query path. When performing requests, the version number
                         prefix will be automatically perpended if required.

        :param Mapping content: The API query body which will be encoded as JSON.

        :param APIConfig config: The API configuration.  If set to ``None``, it will
                                 use the global configuration which is read from the
                                 environment variables.
        '''
        self.config = config if config else get_config()
        self.method = method
        if path.startswith('/'):
            path = path[1:]
        self.path = path
        self.date = None
        self.headers = CIMultiDict([
            ('User-Agent', self.config.user_agent),
            ('X-BackendAI-Version', self.config.version),
        ])
        self.content = content if content is not None else b''
        self.reporthook = reporthook

    @property
    def content(self) -> Union[bytes, bytearray, None]:
        '''
        Retrieves the content in the original form.
        Private codes should NOT use this as it incurs duplicate
        encoding/decoding.
        '''
        if self._content is None:
            raise ValueError('content is not set.')
        if self.content_type == 'application/octet-stream':
            return self._content
        elif self.content_type == 'application/json':
            return json.loads(self._content.decode('utf-8'),
                              object_pairs_hook=OrderedDict)
        elif self.content_type == 'text/plain':
            return self._content.decode('utf-8')
        elif self.content_type == 'multipart/form-data':
            return self._content
        else:
            raise RuntimeError('should not reach here')  # pragma: no cover

    @content.setter
    def content(self, value: Union[bytes, bytearray,
                                   Mapping[str, Any],
                                   Sequence[Any],
                                   None]):
        '''
        Sets the content of the request.
        Depending on the type of content, it automatically sets appropriate
        headers such as content-type and content-length.
        '''
        if isinstance(value, (bytes, bytearray)):
            self.content_type = 'application/octet-stream'
            self._content = value
            self.headers['Content-Type'] = self.content_type
            self.headers['Content-Length'] = str(len(self._content))
        elif isinstance(value, str):
            self.content_type = 'text/plain'
            self._content = value.encode('utf-8')
            self.headers['Content-Type'] = self.content_type
            self.headers['Content-Length'] = str(len(self._content))
        elif isinstance(value, (dict, OrderedDict)):
            self.content_type = 'application/json'
            self._content = json.dumps(value).encode('utf-8')
            self.headers['Content-Type'] = self.content_type
            self.headers['Content-Length'] = str(len(self._content))
        elif isinstance(value, (list, tuple)):
            self.content_type = 'multipart/form-data'
            self._content = value
            # Let the http client library decide the header values.
            # (e.g., message boundaries)
            if 'Content-Length' in self.headers:
                del self.headers['Content-Length']
            if 'Content-Type' in self.headers:
                del self.headers['Content-Type']
        else:
            raise TypeError('Unsupported content value type.')

    def _sign(self, access_key=None, secret_key=None, hash_type=None):
        '''
        Calculates the signature of the given request and adds the
        Authorization HTTP header.
        It should be called at the very end of request preparation and before
        sending the request to the server.
        '''
        if access_key is None:
            access_key = self.config.access_key
        if secret_key is None:
            secret_key = self.config.secret_key
        if hash_type is None:
            hash_type = self.config.hash_type
        hdrs, _ = generate_signature(
            self.method, self.config.version, self.config.endpoint,
            self.date, self.path, self.content_type, self._content,
            access_key, secret_key, hash_type)
        self.headers.update(hdrs)

    def build_url(self):
        base_url = self.config.endpoint.path.rstrip('/')
        major_ver = self.config.version.split('.', 1)[0]
        query_path = self.path.lstrip('/') if len(self.path) > 0 else ''
        path = '{0}/{1}/{2}'.format(base_url, major_ver, query_path)
        canonical_url = self.config.endpoint.with_path(path)
        return str(canonical_url)

    # TODO: attach rate-limit information

    def send(self, *args, loop=None, **kwargs):
        '''
        Sends the request to the server.
        '''
        global _worker_thread
        if _worker_thread is None:
            _worker_thread = _SyncWorkerThread()
            _worker_thread.start()
        _worker_thread.work_queue.put(self.asend(*args, **kwargs))
        result = _worker_thread.done_queue.get()
        _worker_thread.done_queue.task_done()
        if isinstance(result, Exception):
            raise result
        return result

    async def asend(self, *, sess=None, timeout=None):
        '''
        Sends the request to the server.

        This method is a coroutine.
        '''
        assert self.method in self._allowed_methods
        self.date = datetime.now(tzutc())
        self.headers['Date'] = self.date.isoformat()
        if sess is None:
            sess = aiohttp.ClientSession()
        else:
            assert isinstance(sess, aiohttp.ClientSession)
        try:
            async with sess:
                if self.content_type == 'multipart/form-data':
                    data = aiohttp.FormData()
                    for f in self._content:
                        data.add_field(f.name,
                                       f.file,
                                       filename=f.filename,
                                       content_type=f.content_type)
                    assert data.is_multipart
                else:
                    data = self._content
                self._sign()
                async with _timeout(timeout):
                    rqst_ctx = sess.request(
                        self.method,
                        self.build_url(),
                        data=data,
                        headers=self.headers)
                    async with rqst_ctx as resp:
                        body = await resp.read()
                        return Response(resp.status, resp.reason,
                                        body=body,
                                        content_type=resp.content_type,
                                        charset=resp.charset)
        except (asyncio.CancelledError, asyncio.TimeoutError):
            # These exceptions must be bubbled up.
            raise
        except aiohttp.ClientError as e:
            msg = 'Request to the API endpoint has failed.\n' \
                  'Check your network connection and/or the server status.'
            raise BackendClientError(msg) from e

    async def connect_websocket(self, sess=None):
        '''
        Creates a WebSocket connection.
        '''
        assert self.method == 'GET'
        self.date = datetime.now(tzutc())
        self.headers['Date'] = self.date.isoformat()
        if sess is None:
            sess = aiohttp.ClientSession()
        else:
            assert isinstance(sess, aiohttp.ClientSession)
        self._sign()
        try:
            ws = await sess.ws_connect(self.build_url(), headers=self.headers)
            return sess, ws
        except (asyncio.CancelledError, asyncio.TimeoutError):
            # These exceptions must be bubbled up.
            raise
        except aiohttp.ClientError as e:
            msg = 'Request to the API endpoint has failed.\n' \
                  'Check your network connection and/or the server status.'
            raise BackendClientError(msg) from e


class Request(BaseRequest):
    pass


class Response:

    __slots__ = ['_status', '_reason',
                 '_content_type', '_content_length', '_charset',
                 '_body']

    def __init__(self, status: int, reason: str, *,
                 body: Union[bytes, bytearray]=b'',
                 content_type='text/plain',
                 content_length=None,
                 charset=None):
        self._status = status
        self._reason = reason
        if not isinstance(body, (bytes, bytearray)):
            raise ValueError('body must be a bytes-like object.')
        self._body = body
        self._content_type = content_type
        self._content_length = content_length
        self._charset = charset if charset else 'utf-8'

        # TODO: include rate-limiting information from headers

    @property
    def status(self):
        return self._status

    @property
    def reason(self):
        return self._reason

    @property
    def content_type(self):
        return self._content_type

    @property
    def content_length(self):
        is_multipart = self._content_type.startswith('multipart/')
        if self._content_length is None and not is_multipart:
            return len(self._body)
        return self._content_length

    @property
    def charset(self):
        return self._charset

    @property
    def content(self) -> bytes:
        return self._body

    def text(self) -> str:
        return self._body.decode(self._charset)

    def json(self, loads=json.loads):
        return loads(self._body.decode(self._charset), object_pairs_hook=OrderedDict)



from pathlib import Path
import sys

from tabulate import tabulate

from . import register_command
from .pretty import print_done, print_fail
from ..exceptions import BackendError
from ..vfolder import VFolder


@register_command
def vfolder(args):
    '''Provides virtual folder operations.'''
    print('Run with -h/--help for usage.')


@vfolder.register_command
def list(args):
    '''List virtual folders that belongs to the current user.'''
    fields = [
        ('Name', 'name'),
        ('ID', 'id'),
    ]
    try:
        resp = VFolder.list()
        if not resp:
            print('There is no virtual folders created yet.')
            return
        rows = (tuple(vf[key] for _, key in fields) for vf in resp)
        hdrs = (display_name for display_name, _ in fields)
        print(tabulate(rows, hdrs))
    except BackendError as e:
        print_fail(str(e))
        sys.exit(1)


@vfolder.register_command
def create(args):
    '''Create a new virtual folder.'''
    try:
        result = VFolder.create(args.name)
        print('Virtual folder "{0}" is created.'.format(result['name']))
    except BackendError as e:
        print_fail(str(e))
        sys.exit(1)


create.add_argument('name', type=str, help='The name of a virtual folder.')


@vfolder.register_command
def delete(args):
    '''Delete the given virtual folder. This operation is irreversible!'''
    try:
        VFolder(args.name).delete()
    except BackendError as e:
        print_fail(str(e))
        sys.exit(1)


delete.add_argument('name', type=str, help='The name of a virtual folder.')


@vfolder.register_command
def info(args):
    '''Show the information of the given virtual folder.'''
    try:
        result = VFolder(args.name).info()
        print('Virtual folder "{0}" (ID: {1})'.format(result['name'], result['id']))
        print('- Number of files: {0}'.format(result['numFiles']))
    except BackendError as e:
        print_fail(str(e))
        sys.exit(1)


info.add_argument('name', type=str, help='The name of a virtual folder.')


@vfolder.register_command
def upload(args):
    '''Upload a file to the virtual folder from the current working directory.'''
    try:
        VFolder(args.name).upload(args.filenames, show_progress=True)
        print_done('Done.')
    except BackendError as e:
        print_fail(str(e))
        sys.exit(1)


upload.add_argument('name', type=str, help='The name of a virtual folder.')
upload.add_argument('filenames', type=Path, nargs='+',
                    help='Paths of the files to be uploaded.')


@vfolder.register_command
def download(args):
    '''Download a file from the virtual folder to the current working directory.'''
    try:
        VFolder(args.name).download(args.filenames)
        print_done('Done.')
    except BackendError as e:
        print_fail(str(e))
        sys.exit(1)


download.add_argument('name', type=str, help='The name of a virtual folder.')
download.add_argument('filenames', type=Path, nargs='+',
                      help='Paths of the files to be uploaded.')
